@@ -10,6 +10,7 @@
 """
 This module contains classes for XML Schema attributes and attribute groups.
 """
+from copy import copy as _copy
 from decimal import Decimal
 from elementpath.datatypes import AbstractDateTime, Duration, AbstractBinary
 from typing import cast, Any, Callable, Union, Dict, List, Optional, \
@@ -26,7 +27,7 @@ from ..translation import gettext as _
 from ..helpers import get_namespace, get_qname
 
 from .exceptions import XMLSchemaValidationError
-from .xsdbase import XsdComponent, ValidationMixin
+from .xsdbase import XsdComponent, XsdAnnotation, ValidationMixin
 from .simple_types import XsdSimpleType
 from .wildcards import XsdAnyAttribute
 
@@ -416,7 +417,7 @@ class XsdAttributeGroup(
             elif child.tag == XSD_ANY_ATTRIBUTE:
                 any_attribute = self.schema.xsd_any_attribute_class(child, self.schema, self)
                 if None in attributes:
-                    attributes[None] = attr = attributes[None].copy()
+                    attributes[None] = attr = _copy(attributes[None])
                     assert isinstance(attr, XsdAnyAttribute)
                     attr.intersection(any_attribute)
                 else:
@@ -486,7 +487,7 @@ class XsdAttributeGroup(
                                         self.parse_error(msg.format(name))
                                 else:
                                     assert isinstance(base_attr, XsdAnyAttribute)
-                                    attributes[None] = attr = attributes[None].copy()
+                                    attributes[None] = attr = _copy(attributes[None])
                                     assert isinstance(attr, XsdAnyAttribute)
                                     attr.intersection(base_attr)
 
@@ -506,7 +507,7 @@ class XsdAttributeGroup(
                 if name not in self.base_attributes:
                     if self.derivation != 'restriction':
                         continue
-                    elif wildcard is None or not wildcard.is_matching(name, self.default_namespace):
+                    elif wildcard is None or not wildcard.is_matching(name):
                         msg = _("Unexpected attribute {!r} in restriction")
                         self.parse_error(msg.format(name))
                     continue
@@ -592,7 +593,7 @@ class XsdAttributeGroup(
         self._attribute_group.update(attributes)
         if None in self._attribute_group and None not in attributes \
                 and self.derivation == 'restriction':
-            wildcard = self._attribute_group[None].copy()
+            wildcard = _copy(self._attribute_group[None])
             wildcard.namespace = wildcard.not_namespace = wildcard.not_qname = ()
             self._attribute_group[None] = wildcard
 
@@ -613,6 +614,12 @@ class XsdAttributeGroup(
     def built(self) -> bool:
         return True
 
+    @property
+    def annotation(self) -> Optional['XsdAnnotation']:
+        if self.parent is not None and '_annotation' not in self.__dict__:
+            self._annotation = None
+        return super().annotation
+
     def parse_error(self, error: Union[str, Exception],
                     elem: Optional[ElementType] = None,
                     validation: Optional[str] = None) -> None:
