@@ -13,17 +13,20 @@ This module contains classes for other XML Schema identity constraints.
 import re
 import math
 from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Pattern, \
-    Tuple, Union, Counter
+    Set, Tuple, Union, Counter
 from elementpath import XPath2Parser, ElementPathError, XPathToken, XPathContext, \
-    translate_pattern, datatypes
+    ElementNode, translate_pattern, datatypes
 
 from ..exceptions import XMLSchemaTypeError, XMLSchemaValueError
 from ..names import XSD_QNAME, XSD_UNIQUE, XSD_KEY, XSD_KEYREF, XSD_SELECTOR, XSD_FIELD
+from ..translation import gettext as _
 from ..helpers import get_qname, get_extended_qname
 from ..aliases import ElementType, SchemaType, NamespacesType, AtomicValueType
-from ..xpath import iter_schema_nodes
+from .exceptions import XMLSchemaNotBuiltError
 from .xsdbase import XsdComponent
 from .attributes import XsdAttribute
+from .wildcards import XsdAnyElement
+from . import elements
 
 if TYPE_CHECKING:
     from .elements import XsdElement
@@ -44,11 +47,22 @@ XSD_IDENTITY_XPATH_SYMBOLS = frozenset((
 ))
 
 
-# XSD identities use a restricted parser and a context for iterate element
-# references. The XMLSchemaProxy is not used for the specific selection of
-# fields and elements and the XSD fields are got at first validation run.
-class IdentityXPathContext(XPathContext):
-    _iter_nodes = staticmethod(iter_schema_nodes)
+# XSD identities use a restricted XPath 2.0 parser. The XMLSchemaProxy is
+# not used for the specific selection of fields and elements and the XSD
+# fields are collected at first validation run.
+
+def iter_root_elements(token: XPathToken) -> Iterator[XPathToken]:
+    if token.symbol in ('(name)', ':', '*', '.'):
+        yield token
+    elif token.symbol in ('//', '/'):
+        yield from iter_root_elements(token[0])
+        for tk in token[1].iter():
+            if tk.symbol == '|':
+                yield from iter_root_elements(tk[1])
+                break
+    elif token.symbol in '|':
+        for tk in token:
+            yield from iter_root_elements(tk)
 
 
 class IdentityXPathParser(XPath2Parser):
@@ -71,8 +85,8 @@ class XsdSelector(XsdComponent):
         lazy_quantifiers=False,
         anchors=False
     )
-    token = None   # type: XPathToken
-    parser = None  # type: IdentityXPathParser
+    token: Optional[XPathToken] = None
+    parser: Optional[IdentityXPathParser] = None
 
     def __init__(self, elem: ElementType, schema: SchemaType,
                  parent: Optional['XsdIdentity']) -> None:
@@ -82,7 +96,7 @@ class XsdSelector(XsdComponent):
         try:
             self.path = self.elem.attrib['xpath']
         except KeyError:
-            self.parse_error("'xpath' attribute required")
+            self.parse_error(_("'xpath' attribute required"))
             self.path = '*'
         else:
             path = self.path.replace(' ', '')
@@ -94,7 +108,7 @@ class XsdSelector(XsdComponent):
                 _match = self.pattern.match(path)  # type: ignore[union-attr]
 
             if not _match:
-                msg = "invalid XPath expression for an {}"
+                msg = _("invalid XPath expression for an {}")
                 self.parse_error(msg.format(self.__class__.__name__))
 
         # XSD 1.1 xpathDefaultNamespace attribute
@@ -126,7 +140,6 @@ class XsdSelector(XsdComponent):
 
     @property
     def target_namespace(self) -> str:
-        # TODO: implement a property in elementpath for getting XPath token's namespace
         if self.token is None:
             pass  # xpathDefaultNamespace="##targetNamespace"
         elif self.token.symbol == ':':
@@ -164,11 +177,12 @@ class XsdIdentity(XsdComponent):
     parent: 'XsdElement'
     ref: Optional['XsdIdentity']
 
-    selector = None  # type: XsdSelector
-    fields = ()      # type: Union[Tuple[()], List[XsdFieldSelector]]
+    selector: Optional[XsdSelector] = None
+    fields: Union[Tuple[()], List[XsdFieldSelector]] = ()
 
     # XSD elements bound by selector (for speed-up and for lazy mode)
     elements: Union[Tuple[()], Dict['XsdElement', Optional[IdentityCounterType]]] = ()
+    root_elements: Union[Tuple[()], Set['XsdElement']] = ()
 
     def __init__(self, elem: ElementType, schema: SchemaType,
                  parent: Optional['XsdElement']) -> None:
@@ -178,7 +192,7 @@ class XsdIdentity(XsdComponent):
         try:
             self.name = get_qname(self.target_namespace, self.elem.attrib['name'])
         except KeyError:
-            self.parse_error("missing required attribute 'name'")
+            self.parse_error(_("missing required attribute 'name'"))
             self.name = ''
 
         for child in self.elem:
@@ -186,7 +200,7 @@ class XsdIdentity(XsdComponent):
                 self.selector = XsdSelector(child, self.schema, self)
                 break
         else:
-            self.parse_error("missing 'selector' declaration.")
+            self.parse_error(_("missing 'selector' declaration"))
 
         self.fields = []
         for child in self.elem:
@@ -198,68 +212,100 @@ class XsdIdentity(XsdComponent):
             try:
                 ref = self.maps.identities[self.name]
             except KeyError:
-                self.parse_error("unknown identity constraint {!r}".format(self.name))
+                msg = _("unknown identity constraint {!r}")
+                self.parse_error(msg.format(self.name))
                 return
             else:
                 if not isinstance(ref, self.__class__):
-                    self.parse_error("attribute 'ref' points to a different kind constraint")
+                    msg = _("attribute 'ref' points to a different kind constraint")
+                    self.parse_error(msg)
                 self.selector = ref.selector
                 self.fields = ref.fields
                 self.ref = ref
 
-        context = IdentityXPathContext(self.schema, item=self.parent)  # type: ignore
+        if self.selector is None:
+            return  # Do not raise, already found by meta-schema validation.
+        elif self.selector.token is None:
+            raise XMLSchemaNotBuiltError(self, "identity selector is not built")
 
+        context = XPathContext(self.schema.xpath_node, item=self.parent.xpath_node)
         self.elements = {}
-        try:
-            for e in self.selector.token.select_results(context):
-                if not isinstance(e, XsdComponent) or isinstance(e, XsdAttribute):
-                    self.parse_error("selector xpath expression can only select elements")
-                elif e.name is not None:
-                    self.elements[e] = None  # type: ignore[index]
-        except AttributeError:
-            pass
-        else:
-            if not self.elements:
-                # Try to detect target XSD elements extracting QNames
-                # of the leaf elements from the XPath expression and
-                # use them to match global elements.
 
-                qname: Any
-                for qname in self.selector.token.iter_leaf_elements():
-                    xsd_element = self.maps.elements.get(
-                        get_extended_qname(qname, self.namespaces)
-                    )
-                    if xsd_element is not None and \
-                            not isinstance(xsd_element, tuple) and \
-                            xsd_element not in self.elements:
-                        self.elements[xsd_element] = None
+        for e in self.selector.token.select_results(context):
+            if not isinstance(e, (elements.XsdElement, XsdAnyElement)):
+                msg = _("selector xpath expression can only select elements")
+                self.parse_error(msg)
+            elif e.name is not None:
+                if e.ref is not None:
+                    e = e.ref
+                self.elements[e] = None  # XSD fields must be added during validation
+                e.selected_by.add(self)
+
+        if not self.elements:
+            # Try to detect target XSD elements extracting QNames
+            # of the leaf elements from the XPath expression and
+            # use them to match global elements.
+
+            qname: Any
+            for qname in self.selector.token.iter_leaf_elements():
+                xsd_element = self.maps.elements.get(
+                    get_extended_qname(qname, self.namespaces)
+                )
+                if xsd_element is not None and \
+                        not isinstance(xsd_element, tuple) and \
+                        xsd_element not in self.elements:
+                    if xsd_element.ref is not None:
+                        xsd_element = xsd_element.ref
+
+                    self.elements[xsd_element] = None
+                    xsd_element.selected_by.add(self)
+
+        self.root_elements = set()
+        for token in iter_root_elements(self.selector.token):
+            context = XPathContext(self.schema.xpath_node, item=self.parent.xpath_node)
+            for e in token.select_results(context):
+                if isinstance(e, elements.XsdElement):
+                    self.root_elements.add(e)
 
     @property
     def built(self) -> bool:
         return not isinstance(self.elements, tuple)
 
-    def get_fields(self, elem: Union[ElementType, 'XsdElement'],
+    def get_fields(self, element_node: ElementNode,
                    namespaces: Optional[NamespacesType] = None,
                    decoders: Optional[Tuple[XsdAttribute, ...]] = None) -> IdentityCounterType:
         """
         Get fields for a schema or instance context element.
 
-        :param elem: an Element or an XsdElement
+        :param element_node: an Element or an XsdElement
         :param namespaces: is an optional mapping from namespace prefix to URI.
         :param decoders: context schema fields decoders.
         :return: a tuple with field values. An empty field is replaced by `None`.
         """
         fields: List[IdentityFieldItemType] = []
 
-        if not isinstance(elem, XsdComponent):
-            context_class = XPathContext
-        else:
-            context_class = IdentityXPathContext
+        def append_fields() -> None:
+            if isinstance(value, list):
+                fields.append(tuple(value))
+            elif isinstance(value, bool):
+                fields.append((value, bool))
+            elif not isinstance(value, float):
+                fields.append(value)
+            elif math.isnan(value):
+                fields.append(('nan', float))
+            else:
+                fields.append((value, float))
 
         result: Any
         value: Union[AtomicValueType, None]
+
         for k, field in enumerate(self.fields):
-            result = field.token.get_results(context_class(elem))  # type: ignore
+            if field.token is None:
+                msg = f"identity field {field} is not built"
+                raise XMLSchemaNotBuiltError(self, msg)
+
+            context = XPathContext(element_node)
+            result = field.token.get_results(context)
 
             if not result:
                 if decoders is not None and decoders[k] is not None:
@@ -268,26 +314,16 @@ class XsdIdentity(XsdComponent):
                         if decoders[k].type.root_type.name == XSD_QNAME:
                             value = get_extended_qname(value, namespaces)
 
-                        if isinstance(value, list):
-                            fields.append(tuple(value))
-                        elif isinstance(value, bool):
-                            fields.append((value, bool))
-                        elif not isinstance(value, float):
-                            fields.append(value)
-                        elif math.isnan(value):
-                            fields.append(('nan', float))
-                        else:
-                            fields.append((value, float))
-
+                        append_fields()
                         continue
 
-                if not isinstance(self, XsdKey) or 'ref' in elem.attrib and \
+                if not isinstance(self, XsdKey) or 'ref' in element_node.elem.attrib and \
                         self.schema.meta_schema is None and self.schema.XSD_VERSION != '1.0':
                     fields.append(None)
                 elif field.target_namespace not in self.maps.namespaces:
                     fields.append(None)
                 else:
-                    msg = "missing key field {!r} for {!r}"
+                    msg = _("missing key field {0!r} for {1!r}")
                     raise XMLSchemaValueError(msg.format(field.path, self))
 
             elif len(result) == 1:
@@ -295,7 +331,8 @@ class XsdIdentity(XsdComponent):
                     fields.append(result[0])
                 else:
                     if decoders[k].type.content_type_label not in ('simple', 'mixed'):
-                        raise XMLSchemaTypeError("%r field doesn't have a simple type!" % field)
+                        msg = _("%r field doesn't have a simple type!")
+                        raise XMLSchemaTypeError(msg % field)
 
                     value = decoders[k].data_value(result[0])
                     if decoders[k].type.root_type.name == XSD_QNAME:
@@ -304,23 +341,15 @@ class XsdIdentity(XsdComponent):
                         elif isinstance(value, datatypes.QName):
                             value = value.expanded_name
 
-                    if isinstance(value, list):
-                        fields.append(tuple(value))
-                    elif isinstance(value, bool):
-                        fields.append((value, bool))
-                    elif not isinstance(value, float):
-                        fields.append(value)
-                    elif math.isnan(value):
-                        fields.append(('nan', float))
-                    else:
-                        fields.append((value, float))
+                    append_fields()
             else:
-                raise XMLSchemaValueError("%r field selects multiple values!" % field)
+                msg = _("%r field selects multiple values!")
+                raise XMLSchemaValueError(msg % field)
 
         return tuple(fields)
 
-    def get_counter(self, enabled: bool = True) -> 'IdentityCounter':
-        return IdentityCounter(self, enabled)
+    def get_counter(self, elem: ElementType) -> 'IdentityCounter':
+        return IdentityCounter(self, elem)
 
 
 class XsdUnique(XsdIdentity):
@@ -348,7 +377,7 @@ class XsdKeyref(XsdIdentity):
             self.refer = self.schema.resolve_qname(self.elem.attrib['refer'])
         except (KeyError, ValueError, RuntimeError) as err:
             if 'refer' not in self.elem.attrib:
-                self.parse_error("missing required attribute 'refer'")
+                self.parse_error(_("missing required attribute 'refer'"))
             else:
                 self.parse_error(err)
 
@@ -363,20 +392,29 @@ class XsdKeyref(XsdIdentity):
         if self.refer is None:
             return  # attribute or key/unique identity constraint missing
         elif isinstance(self.refer, str):
-            refer = self.parent.identities.get(self.refer)
+            refer: Optional[XsdIdentity]
+            for refer in self.parent.identities:
+                if refer.name == self.refer:
+                    break
+            else:
+                refer = None
+
             if refer is not None and refer.ref is None:
                 self.refer = refer  # type: ignore[assignment]
             else:
                 try:
                     self.refer = self.maps.identities[self.refer]  # type: ignore[assignment]
                 except KeyError:
-                    self.parse_error("key/unique identity constraint %r is missing" % self.refer)
+                    msg = _("key/unique identity constraint %r is missing")
+                    self.parse_error(msg % self.refer)
                     return
 
         if not isinstance(self.refer, (XsdKey, XsdUnique)):
-            self.parse_error("reference to a non key/unique identity constraint %r" % self.refer)
+            msg = _("reference to a non key/unique identity constraint %r")
+            self.parse_error(msg % self.refer)
         elif len(self.refer.fields) != len(self.fields):
-            self.parse_error("field cardinality mismatch between %r and %r" % (self, self.refer))
+            msg = _("field cardinality mismatch between {0!r} and {1!r}")
+            self.parse_error(msg.format(self, self.refer))
         elif self.parent is not self.refer.parent:
             refer_path = self.refer.parent.get_path(ancestor=self.parent)
             if refer_path is None:
@@ -400,8 +438,8 @@ class XsdKeyref(XsdIdentity):
     def built(self) -> bool:
         return not isinstance(self.elements, tuple) and isinstance(self.refer, XsdIdentity)
 
-    def get_counter(self, enabled: bool = True) -> 'KeyrefCounter':
-        return KeyrefCounter(self, enabled)
+    def get_counter(self, elem: ElementType) -> 'KeyrefCounter':
+        return KeyrefCounter(self, elem)
 
 
 class Xsd11Unique(XsdUnique):
@@ -430,22 +468,24 @@ class Xsd11Keyref(XsdKeyref):
 
 class IdentityCounter:
 
-    def __init__(self, identity: XsdIdentity, enabled: bool = True) -> None:
+    def __init__(self, identity: XsdIdentity, elem: ElementType) -> None:
         self.counter: Counter[IdentityCounterType] = Counter[IdentityCounterType]()
         self.identity = identity
-        self.enabled = enabled
+        self.elem = elem
+        self.enabled = True
 
     def __repr__(self) -> str:
         return "%s%r" % (self.__class__.__name__[:-7], self.counter)
 
-    def clear(self) -> None:
+    def reset(self, elem: ElementType) -> None:
         self.counter.clear()
+        self.elem = elem
         self.enabled = True
 
     def increase(self, fields: IdentityCounterType) -> None:
         self.counter[fields] += 1
         if self.counter[fields] == 2:
-            msg = "duplicated value {!r} for {!r}"
+            msg = _("duplicated value {0!r} for {1!r}")
             raise XMLSchemaValueError(msg.format(fields, self.identity))
 
 
